"""
Config system based on functions

The config system consists of two ends:
* a function can be declared config, which means it is expected
  to be replaced by an actual implementation
* a way to set these functions to actual implementations

The "declaring" side uses a decorator (`@declare_config()`) to declare
a function a configuration function.

It can use environment variables and actual function implementations
(decorated with `@set_config('module.name')` to replace the config function.
"""
import functools
import logging
import os
import sys
from typing import Callable, Tuple, Dict, Generator

__all__ = ['declare_config', 'set_config', 'patch', 'wrap',
           'add_config_from_local_setup_py', 'add_config_from_environment']

log = logging.getLogger(__name__)

__CONFIG_REGISTRY: Dict[str, Tuple[Callable, bool]] = {}
__ORIG_API_REGISTRY: Dict[str, Callable] = {}


def _get_full_name(func: Callable) -> str:
    return (func.__module__ or '<no_module>') + '.' + func.__name__


def declare_config(config_name=None):
    """Decorator for config or API functions which can be replaced by downstream packages

    Args:
        config_name: str, default: None
            name of the config function. If None, the function is registered under
            the name 'modulename.funcname'.

    Returns: The decorated function, which looks up a replacing function in the config system
    """
    outer_config_name = config_name

    def _replaceable(func):
        config_name = (outer_config_name if outer_config_name
                       else _get_full_name(func))
        log.debug("Registered new replaceable function '%s'", config_name)
        __ORIG_API_REGISTRY[config_name] = func

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if config_name in __CONFIG_REGISTRY:
                replacement_func, include_original_function = __CONFIG_REGISTRY[config_name]
                if include_original_function:
                    kwargs['original_function'] = func

                return replacement_func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

        setattr(wrapper, 'config_name', config_name)
        return wrapper

    return _replaceable


def set_config(config_name: str, include_original_function=False, function: Callable = None):
    """Sets a config or API function

    Can be used as a decorator:
    >>> @set_config('mara_without_function_pointer')
    >>> def replacement(...):
    >>>     pass

    or as a function:
    >>> def replacement(...):
    >>>     pass
    >>> if should_be_replaced:
    >>>     set_config('mara_without_function_pointer', replacement)

    Args:
        config_name: str
            name of the to be replaced config function
        include_original_function: Boolean, default: False
            whether or not to pass in the original function to this decorated function
        function: Callable, default: None
            The function which replaces the declared config function (only needed when used as a function,
            do not set when used as decorator)

    Returns: The replacing function (undecorated but registered in the config system)
    """
    if function is None:
        # usage as decorator
        def submitting_decorator(func):
            set_config(config_name=config_name, include_original_function=include_original_function, function=func)
            return func

        return submitting_decorator
    else:
        assert callable(function), f"New function for '{config_name}' is not callable: {type(function)}"
        if config_name in __CONFIG_REGISTRY:
            orig_replacement, _ = __CONFIG_REGISTRY[config_name]
            log.warning("Replacing already replaced function for '%s': %s.%s",
                        config_name, orig_replacement.__module__, orig_replacement.__name__)
        __CONFIG_REGISTRY[config_name] = (function, include_original_function)
        log.debug("Replacing function '%s' with %s.%s", config_name, function.__module__, function.__name__)


# Monkey patching: For replacing functionality which does not fit your need
# See See https://en.wikipedia.org/wiki/Monkey_patch

def _monkey_patch(original_function: Callable, include_original_function: bool) -> Callable:
    """Replace or wrap functions in other modules"""

    # Make the original function a @config decorated function:
    config_func = declare_config()(original_function)
    config_name = _get_full_name(original_function)

    # replace it
    orig = getattr(sys.modules[original_function.__module__], original_function.__name__)
    setattr(sys.modules[original_function.__module__], original_function.__name__, config_func)
    new = getattr(sys.modules[original_function.__module__], original_function.__name__)
    assert orig != new

    # basically decorate the new function with @set_config
    def decorator(new_function):
        # the config system includes the function as a kwarg, the wrap wants it as a first argument
        @functools.wraps(new_function)
        def f(*args, **kwargs):
            if 'original_function' in kwargs:
                first_arg = kwargs.pop('original_function')
                new_args = [first_arg] + list(args)
            else:
                new_args = args
            return new_function(*new_args, **kwargs)

        new_func = f
        set_config(config_name, include_original_function=include_original_function, function=new_func)
        return new_function

    return decorator


def patch(original_function: Callable) -> Callable:
    """A decorator for replacing a function in another module

    Internally, replaces the original function with a `@declare_config`
    decorated functions and decorates the the new function with `@set_config`.

        Example:
    >>> # in some_package.some_module (not decorated with @declare_config!):
    ... def some_function(x):
    ...     return x + 1

    >>> some_package.some_module.some_function(1)
    2

    >>> @patch(some_package.some_module.some_function)
    ... def new_function(x):
    ...      return x + 2

    >>> some_package.some_module.some_function(1)
    3

    >>> # equivalent:
    >>> patch(some_package.some_module.some_function)(lambda x: x + 2)

    Args:
        original_function: The function or method to patch

    Returns: The replaced function
    """
    return _monkey_patch(original_function, False)


def wrap(original_function: Callable) -> Callable:
    """A decorator for wrapping a function in another module

    Internally, replaces the original function with a `@declare_config`
    decorated functions and decorates the the new function with `@set_config`.

        Example:
    >>> # in some_package.some_module:
    ... def some_function(x):
    ...     return x + 1

    >>> some_package.some_module.some_function(1)
    2

    >>> @wrap(some_package.some_module.some_function)
    ... def new_function(original_function, x):
    ...      return original_function(x) + 1

    >>> some_package.some_module.some_function(1)
    3

    Args:
        original_function: The function or method to wrap

    Returns: The wrapped function
    """
    return _monkey_patch(original_function, True)


def _reset_config():
    """Reset config internal state

    Internal function for testing purpose"""
    for k in list(__CONFIG_REGISTRY.keys()):
        del __CONFIG_REGISTRY[k]
    for k in list(__ORIG_API_REGISTRY.keys()):
        del __ORIG_API_REGISTRY[k]


def get_overwritten_config() -> Generator[Tuple[str, Tuple[Callable, bool]], None, None]:
    """Get all overwritten config"""
    for k, v in __CONFIG_REGISTRY.items():
        yield k, v


def get_declared_config() -> Generator[Tuple[str, Callable], None, None]:
    """Get all declared config"""
    for k, v in __ORIG_API_REGISTRY.items():
        yield k, v


## load config from environment


def _bool(value):
    valid_str = {'true': True, 't': True, '1': True,
                 'false': False, 'f': False, '0': False,
                 }
    lower_value = value.lower()
    try:
        return valid_str[lower_value]
    except:
        raise ValueError('invalid literal for boolean: "%s"' % value)


def _float(input: str) -> float:
    # workaround for getting "1,0" recognized as "1.0"
    val = float(input.replace(',', '.'))
    return val


def add_config_from_environment():
    """Add configuration from the environment

    This only works for config items which return either booleans, numbers (floats), or strings.

    Any environment variable (in lowercase) which starts 'mara_' is turned into
    functions which returns the value. The rest of the environment variable name
    has any '__' replaced by '.'. If the value is a valid float or boolean, it's returned
    as a float/boolean. Otherwise it's returned as a string.

    E.g. the following variable

        MARA_PACKAGENAME__CONFIG_ITEM=y

    is equivalent to the following @set_config call

    >>> @set_config('packagename.config_item')
    >>> def replacement():
    >>>     return 'y'

    """
    from ..config import default_environment_prefix
    prefix = default_environment_prefix().lower()
    loaded = False
    for k in os.environ.keys():
        key = str(k).lower()
        if key.startswith(prefix):
            config_name = key[len(prefix) + 1:].replace('__', '.')
            value = os.environ[k]
            try:
                # try to convert to a value with a more specific type
                value = _float(value)
            except:
                try:
                    value = _bool(value)
                except:
                    # we treat it as a string
                    pass
            loaded = True
            # This needs to bound the value inside the lambda, otherwise the value will get from the outer
            # scope and in that case value will be the last processed environment variable
            # https://stackoverflow.com/a/19837683/1380673
            set_config(config_name, False, lambda bound_value=value: bound_value)
    if loaded:
        log.debug("Loaded some config from environment")

def prepare_import():
    """Try to find the best folder to add to the search path
    """
    path = os.getcwd()
    path = os.path.realpath(path)

    if os.path.exists(os.path.join(path, '__init__.py')):
        # move up until outside package structure (no __init__.py)
        while True:
            path, name = os.path.split(path)

            if not os.path.exists(os.path.join(path, '__init__.py')):
                break

    if sys.path[0] != path:
        sys.path.insert(0, path)


def add_config_from_local_setup_py():
    # apply environment specific settings (not in git repo)
    import importlib
    from ..config import default_app_module
    prepare_import()
    parts = default_app_module().split('.')
    while parts:
        try:
            pos_module = '.'.join(parts) + '.local_setup'
            importlib.import_module(pos_module)
            log.debug("Loaded config from local_setup.py at %s", pos_module)
            return
        except ModuleNotFoundError:
            pass
        parts.pop()
    log.debug("No local_setup.py found.")
    return



"""Default configuration visible in all mara packages"""
import os

from .config_system import declare_config

__all__ = ['debug', 'default_app_module', 'default_environment_prefix']

@declare_config("debug")
def debug():
    """Whether this app should output debug information (Default: False)"""
    return False


@declare_config("app")
def default_app_module():
    """The app module where the app composing functions are defined (Default: 'app.app')

    The only way to set this before this is used is by setting an environment variable $MARA_APP.
    """
    # this is not replaceable in local_setup.py, only '@config' to get it show up in the flask view
    # hack to already get the info before the config system is fully initialized
    return os.environ.get('MARA_APP', 'app.app')

@declare_config("default_environment_prefix")
def default_environment_prefix():
    """The prefix to identify configuration variables in the environment (Default: 'MARA')
    """
    return os.environ.get('MARA_DEFAULT_ENVIRONMENT_PREFIX', 'MARA')



"""App composing and Configuration infrastructure"""
from .config_system import declare_config, set_config
from .app_composing import (call_app_composing_function,
                            register_functionality, get_contributed_functionality,
                            register_functionality_in_all_imported_modules)


# Contributing the config stuff to the FLASK APP (mara_app)
def MARA_FLASK_BLUEPRINTS():
    import mara_config.config_system.view as view
    yield view.mara_config


def MARA_ACL_RESOURCES():
    import mara_config.config_system.view as view
    yield view.acl_resource


def MARA_NAVIGATION_ENTRY_FNS():
    import mara_config.config_system.view as view
    yield view.navigation_entry

# And something to the config system itself
def MARA_CONFIG():
    import mara_config.config
    yield mara_config.config



import logging

import pytest
import sys
from mara_config import declare_config, set_config
from mara_config.config_system import add_config_from_environment, _reset_config


# this dance with orig_something and so on is needed so 'something' is actually decorated by the fixture
# and the config can be cleaned up
def something(argument: str = None) -> str:
    "Test API function"
    return "x"


orig_something = something


def without_args() -> str:
    "Test without arguments"
    return "x"


orig_without_args = without_args


def wraps_test_func():
    return 'x'
def patch_test_func():
    return 'x'


@pytest.fixture()
def setup_config():
    """Setup a clean config and insert a single API"""
    _reset_config()
    logging.root.setLevel(logging.DEBUG)
    global something
    global without_args
    something = declare_config()(orig_something)
    without_args = declare_config()(orig_without_args)
    # global wraps_test_func, patch_test_func
    # wraps_test_func = lambda: 'x'
    # wraps_test_func.__name__ = 'wraps_test_func'
    # patch_test_func = lambda: 'x'
    # patch_test_func.__name__ = 'patch_test_func'
    yield setup_config
    _reset_config()


# use it in every tests in this file
pytestmark = pytest.mark.usefixtures("setup_config")


def test_config_decorator():
    @declare_config()
    def _tester(argument: str = None) -> str:
        return "x"

    # unreplaced
    assert 'x' == _tester()
    assert 'x' == _tester("ABC")


def test_replace_decorator_without_function_pointer():
    # In downstream package which want's to overwrite the API
    @set_config('mara_config.tests.test_config.something')
    def replacement_for_something(argument: str = None) -> str:
        return argument or 'y'

    assert 'y' == something()
    assert 'ABC' == something("ABC")


def test_replace_decorator_with_include_orig_function():
    # In downstream package which want's to overwrite the API
    @set_config('mara_config.tests.test_config.something', include_original_function=True)
    def replacement_for_something(argument: str = None, original_function=None) -> str:
        if original_function:
            val = original_function()
        else:
            val = ''
        return val + (argument or 'y')

    assert 'xy' == something()
    assert 'xABC' == something("ABC")

def _get_func(name):
    return getattr(sys.modules[__name__], name)

def test_wraps():

    before = _get_func('wraps_test_func')
    assert 'x' == _get_func('wraps_test_func')()

    from mara_config.config_system import wrap

    @wrap(_get_func('wraps_test_func'))
    def replacement(original_function) -> str:
        str(original_function)
        val = original_function()
        return val + 'y'
    after = _get_func('wraps_test_func')
    assert before != after
    assert 'xy' == _get_func('wraps_test_func')()

def test_patch():

    assert 'x' == _get_func('patch_test_func')()

    from mara_config.config_system import patch

    @patch(_get_func('patch_test_func'))
    def replacement() -> str:
        return 'y'

    assert 'y' == _get_func('patch_test_func')()

def test_replace_decorator_with_function_pointer():
    # In downstream package which want's to overwrite the API
    @set_config('mara_config.tests.test_config.something', include_original_function=True)
    def replacement_for_something(argument: str = None, original_function=None) -> str:
        assert callable(original_function)
        tmp = original_function(argument)
        return tmp + (argument or 'y')

    assert 'xy' == something()
    assert 'xABC' == something("ABC")


def test_replace_function():
    def replacement_for_something(argument: str = None) -> str:
        return argument or 'y'

    set_config('mara_config.tests.test_config.something', function=replacement_for_something)

    assert 'y' == something()
    assert 'ABC' == something("ABC")


def test_replace_function_with_non_function():
    with pytest.raises(AssertionError):
        set_config('mara_config.tests.test_config.something', function="something")


def test_warn_on_use_replace_decorator_twice(caplog):
    # In downstream package which want's to overwrite the API
    caplog.set_level(logging.INFO)

    @set_config('mara_config.tests.test_config.something')
    def replacement_for_something(argument: str = None) -> str:
        return argument or 'y'

    @set_config('mara_config.tests.test_config.something')
    def replacement_for_something2(argument: str = None) -> str:
        return 'z'

    assert len(caplog.records) == 1
    assert "already replaced" in caplog.records[0].message

    assert 'z' == something()
    assert 'z' == something("ABC")


def test_add_config_from_environment():
    import os
    os.environ['mara_mara_config__tests__test_config__without_args'] = 'y'

    assert 'x' == without_args()

    add_config_from_environment()

    assert 'y' == without_args()



"""Helper functions for displaying the content of the config system"""

import inspect
import itertools
import pprint
import sys

from mara_config import get_contributed_functionality

from . import get_declared_config, get_overwritten_config


class ConfigFunction():
    """A object which holds information about a config function"""

    def __init__(self, name):
        self.config_name = name
        self.set_func = None
        self.declared_func = None
        self.include_parent = False
        self.__initialized = False

    def _build_data(self):
        if self.__initialized:
            return
        sig = inspect.signature(self._value_func)
        self._signature = sig
        self.argument_length = len(sig.parameters)
        self.func_name = self._value_func.__name__
        self.module_name = self._value_func.__module__
        try:
            self.value = self._value_func()
            self.error = None
        except Exception as e:
            self.value = f"[Function returned Error]"
            self.error = (str(e.__class__.__name__), str(e))
        self.__initialized = True

    @property
    def _value_func(self):
        return self.set_func or self.declared_func

    @property
    def doc(self):
        doc = self._value_func.__doc__
        if not doc and self.set_func:
            doc = self.set_func.__doc__
        if not doc and self.declared_func:
            doc = self.declared_func.__doc__
        return doc or ''

    @property
    def func_desc(self):
        self._build_data()
        args = []
        for i in self._signature.parameters.values():
            args.append(i.name)
        arg_spec = ', '.join(args)
        return f'{self.module_name}.{self.func_name}({arg_spec})'

    @property
    def state_desc(self):
        self._build_data()
        state = (('S' if self.set_func else '-') +
                 ('D' if self.declared_func else '-') +
                 ('I' if self.include_parent else '-'))
        return state

    def __repr__(self):
        self._build_data()
        if not self.error:
            return pprint.pformat(self.value)
        else:
            if self.argument_length == 0:
                # for now we only report the exception type
                error = f" ({self.error[0]})"
            else:
                # expected because of missing arguments, so no error
                error = ""
            return f"Function '{self.func_desc}'" + error


class ConfigModule():
    """All config functions in a module"""

    def __init__(self, module, contributed):
        self.module = module
        self.name = module.__name__
        self.doc = module.__doc__
        self.contributed = contributed
        self.config_functions = {}

    def get_or_build_function(self, name):
        if name in self.config_functions:
            return self.config_functions['name']
        cf = ConfigFunction(name)
        self.config_functions[name] = cf
        return cf

    def items(self):
        return sorted(self.config_functions.items())


class ConfigForDisplay():
    """The config as a iterable data structure for displaying"""

    def __init__(self):
        self._contributed = {}
        self._uncontributed = {}

    def add_module(self, module: ConfigModule):
        if module.contributed:
            self._contributed[module.name] = module
        else:
            self._uncontributed[module.name] = module

    def get_or_build_module(self, module_name):
        if module_name in self._contributed:
            return self._contributed[module_name]
        if module_name in self._uncontributed:
            return self._uncontributed[module_name]
        cm = ConfigModule(sys.modules[module_name], False)
        self.add_module(cm)
        return cm

    def get_function(self, key):
        for _, module in self.items():
            if key in module.config_functions:
                return module.config_functions[key]

    def items(self):
        return itertools.chain(sorted(self._contributed.items()), sorted(self._uncontributed.items()))


def get_config_for_display() -> ConfigForDisplay:
    """Returns the whole known config

    The ConfigForDisplay contains ConfigModules which contain
    ConfigFunctions.

    Will load all config modules contributed via `MARA_CONFIG`.
    """

    config = ConfigForDisplay()

    for module, config_module in get_contributed_functionality('MARA_CONFIG'):
        config.add_module(ConfigModule(config_module, contributed=True))

    for key, func in get_declared_config():
        cm = config.get_or_build_module(func.__module__)
        cf = cm.get_or_build_function(key)
        cf.declared_func = func
    for key, v in get_overwritten_config():
        func, include_parent = v
        cf = config.get_function(key)
        if not cf:
            cm = config.get_or_build_module(func.__module__)
            cf = cm.get_or_build_function(key)
        cf.set_func = func
        cf.include_parent = include_parent
    return config


def print_config():
    config = get_config_for_display()
    # Get to longest key length
    max_len = 0
    for _, config_module in config.items():
        for k, _ in config_module.items():
            max_len = max(len(k), max_len)

    format_str = f'%-{max_len}s [%s] -> %-{20-max_len}s'
    print('\nConfig:\n-------\n')

    for module_name, config_module in config.items():
        # print(f'# Module: {module_name}')
        for conf_name, conf_func in config_module.items():
            print(format_str % (conf_name, conf_func.state_desc, str(conf_func)))

    print('\nS: config is set, D: @declare_config was called, I: config value includes parent\n')



"""Infrastructure to add and consume functionality"""
import collections
import copy
import itertools
import logging
import sys
import types
import typing

__all__ = ['register_functionality', 'register_functionality_in_all_imported_modules',
           'get_contributed_functionality', 'call_app_composing_function']

log = logging.getLogger(__name__)

# The main API functionality
_registered_functionality: {str: list} = collections.defaultdict(list)
_registered_modules: [str] = []


def register_functionality(module: types.ModuleType):
    """Registers all declared functionality"""
    if module.__name__ in _registered_modules:
        # already registered
        return
    for attr in dir(module):
        if attr.startswith('MARA_'):
            items = getattr(module, attr)
            assert (callable(items) or isinstance(items, typing.Iterable))
            log.debug("Registered '%s' in module '%s'", attr, module.__name__)
            _registered_functionality[attr].append((module, items))
            _registered_modules.append(module.__name__)


def get_contributed_functionality(name: str) -> typing.Iterable:
    """Gets the contributed functionality for one MARA_ variable"""
    all_items = _registered_functionality[name]
    for module, items in all_items:
        if callable(items):
            # a generator
            try:
                yield from zip(itertools.repeat(module), items())
            except Exception as e:
                # if we get a problem, just go over it to not stall the whole app
                log.exception(e)
        else:
            # lists and so on
            try:
                yield from zip(itertools.repeat(module), items)
            except Exception as e:
                log.exception(e)


def register_functionality_in_all_imported_modules():
    """ Imports all contributed mara functionality from any imported modules

    For compatibility with earlier versions of mara_app"""
    for name, module in copy.copy(sys.modules).items():
        register_functionality(module)


def call_app_composing_function():
    """Finds and calls the app composing function

    Tries to import the Module returned by the
    `default_app_module()` (default: app.app) and then call the
    `compose_mara_app()` function in that module, if it exists.
    """
    import importlib
    from .config import default_app_module
    from .config_system import prepare_import
    app_module_name = default_app_module()
    prepare_import()
    try:
        app = importlib.import_module(app_module_name)
    except ModuleNotFoundError:
        msg = f"MARA_APP ({app_module_name}) is not an importable module."
        raise RuntimeError(msg) from None
    if not hasattr(app, 'compose_mara_app'):
        log.error("MARA_APP (%s) has no 'compose_mara_app() function.", app_module_name)
        # this is still recoverable as someone might have just did all the registering on import
        return
    compose_mara_app = getattr(app, 'compose_mara_app')
    log.debug("About to call '%s.compose_mara_app()'", app_module_name)
    try:
        compose_mara_app()
    except BaseException as e:
        msg = "Calling '%s.compose_mara_app()' resulted in an exception"
        raise RuntimeError(msg) from e
    log.debug("Finished '%s.compose_mara_app()'", app_module_name)
    return



"""Mara config views"""

import html
import logging

import flask
from mara_page import acl
from mara_page import navigation, response, _, bootstrap

from mara_config.config_system.config_display import get_config_for_display

log = logging.getLogger(__name__)

mara_config = flask.Blueprint('mara_config', __name__, url_prefix='/config_system', static_folder='static')

acl_resource = acl.AclResource('Configuration')


@mara_config.route('/')
@acl.require_permission(acl_resource)
def configuration_page():
    config = get_config_for_display()
    current_user_has_permission = acl.current_user_has_permission(acl_resource)

    return response.Response(
        html=[(bootstrap.card(id=module_name,
            header_left=html.escape(module_name),
            body=[_.p[_.em[html.escape(str(module_content.doc))]],
                  bootstrap.table(
                      [],
                      [_.tr[
                           _.td[_.tt[html.escape(config_name).replace('_', '_<wbr/>')],
                                [_.br, ' ⟵ ', _.tt[html.escape(config_function.func_desc)
                                    .replace('.', '<wbr/>.').replace('_', '_<wbr/>')]]
                                if config_function.set_func else ''],
                           _.td[_.em[html.escape(config_function.doc)]],
                           _.td[
                               _.pre[html.escape(str(config_function))]
                               if current_user_has_permission
                               else acl.inline_permission_denied_message()
                           ]] for config_name, config_function in module_content.items()])
                  ]) if module_content.config_functions else '')
              for module_name, module_content in config.items()],
        title='Mara Configuration')



def navigation_entry():
    return navigation.NavigationEntry('Configuration (via config system)',
                                      uri_fn=lambda: flask.url_for('mara_config.configuration_page'),
                                      icon='cogs', rank=100)



from setuptools import setup, find_packages

def get_long_description():
    with open('README.md') as f:
        return f.read()

setup(
    name='mara-config',
    version='0.1.1',

    description="Mara app composing and configuration infrastructure.",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',



    extras_require={
        'test': ['pytest',
                 'flask>=0.12', 'mara_page' # config views
        ],
    },

    dependency_links=[
        'git+https://github.com/mara/mara-page.git@1.2.3#egg=mara-page-1.2.3',
    ],

    packages=find_packages(),

    author='Mara contributors',
    license='MIT',

)
